import locale
import logging
from typing import Any, Optional, Tuple

import numpy as np
from numpy import ndarray
from pandas import Series, to_numeric
from pandas.api.types import is_numeric_dtype

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class IntegerType(BaseType):
    weight = 7

    def validate(self, data: ndarray) -> Optional[InferredField]:
        for value in data:
            try:
                self.cast_value(value)
            except Exception:  # noqa
                logger.debug(f"Cannot cast the value '{value}' to {self.name}")
                return None
        return InferredField(inferred_type=self.name)

    @property
    def prefer_column_casting(self) -> bool:
        return True

    def cast_column(self, column: Series, inferred_field: InferredField) -> Series:
        if is_numeric_dtype(column):
            return column
        return to_numeric(column.apply(locale.atoi))

    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        if isinstance(value, str):
            return int(locale.atoi(value))
        return int(value)

    def ignore_cast_for_types(self) -> Tuple[Any]:
        return (np.dtype("int64"),)



import logging
from typing import Any, Optional, Tuple

import numpy as np
from numpy import ndarray
from pandas import Series
from pydantic import BaseModel, EmailStr

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class Model(BaseModel):
    email: EmailStr


class EmailType(BaseType):
    weight = 5

    def validate(self, data: ndarray) -> Optional[InferredField]:
        for value in data:
            try:
                Model(email=value)
            except ValueError:
                logger.debug(f"Cannot cast the value '{value}' to {self.name}")
                return None
        return InferredField(inferred_type=self.name)

    def cast_column(self, column: Series, inferred_field: InferredField) -> Series:
        raise NotImplementedError()

    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        return str(value)

    def ignore_cast_for_types(self) -> Tuple[Any]:
        return (np.dtype("O"),)



from typing import Any, List, Optional

from pydantic import BaseModel


class BaseField(BaseModel):
    field_name: Any
    source_type: str


class InferredField(BaseModel):
    inferred_type: Optional[str] = None
    inferred_pattern: Optional[Any] = None


class Field(InferredField, BaseField):
    pass


class ShmessySchema(BaseModel):
    columns: List[Field]
    infer_duration_ms: int



import logging
import math
from datetime import datetime
from enum import Enum
from typing import Any, Optional, Tuple

import numpy as np
from numpy import ndarray
from pandas import Series, to_datetime

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class TimestampResolution(str, Enum):
    SECONDS = "s"
    MILLISECONDS = "ms"
    NANOSECONDS = "ns"


class UnixTimestampType(BaseType):
    weight = 4
    min_valid_year: int = 1980
    max_valid_year: int = 2100

    resolution: Optional[TimestampResolution] = None

    @staticmethod
    def _unix_timestamp_resolution(value: int) -> TimestampResolution:
        number_of_digits = len(str(int(value)))
        if number_of_digits == 10:
            return TimestampResolution.SECONDS
        if number_of_digits == 13:
            return TimestampResolution.MILLISECONDS
        if number_of_digits == 16:
            return TimestampResolution.NANOSECONDS

    @staticmethod
    def _fix_input_resolution(
        value: int, selected_resolution: TimestampResolution
    ) -> int:
        if selected_resolution == TimestampResolution.SECONDS:
            return int(value)
        if selected_resolution == TimestampResolution.MILLISECONDS:
            return int(int(value) / 1000)
        if selected_resolution == TimestampResolution.NANOSECONDS:
            return int(int(value) / 1000 / 1000)

    def _is_valid_unix_timestamp(self, value: Any) -> bool:
        if self.is_empty_value(value):
            return True

        if isinstance(value, float) and math.isinf(value):
            return False

        if not self.resolution:
            self.resolution = self._unix_timestamp_resolution(int(value))

        if self.resolution:
            parsed_value = datetime.utcfromtimestamp(
                self._fix_input_resolution(value, self.resolution)
            )
            if self.min_valid_year <= parsed_value.year <= self.max_valid_year:
                return True
        return False

    def validate(self, data: ndarray) -> Optional[InferredField]:
        self.resolution = None
        try:
            for value in data:
                if not self._is_valid_unix_timestamp(value):
                    return None

            if not self.resolution:
                return None

            return InferredField(
                inferred_type=self.name, inferred_pattern=self.resolution
            )

        except (ValueError, OSError, OverflowError) as e:
            logger.debug(f"Cannot cast the given data to {self.name}: {e}")
            return None

    @property
    def prefer_column_casting(self) -> bool:
        return True

    def cast_column(self, column: Series, inferred_field: InferredField) -> Series:
        return to_datetime(column, unit=inferred_field.inferred_pattern.value)

    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        if self.is_empty_value(value):
            return None
        return datetime.fromtimestamp(self._fix_input_resolution(value, pattern))

    def ignore_cast_for_types(self) -> Tuple[Any]:
        return (np.dtype("datetime64"),)



from typing import Any, Optional, Tuple

import numpy as np
from numpy import ndarray
from pandas import Series

from ..schema import InferredField
from .base import BaseType


class BooleanType(BaseType):
    weight = 1
    patterns: list[Tuple] = [  # The first member should be the true value
        ("YES", "NO"),
        ("TRUE", "FALSE"),
        ("T", "F"),
        ("Y", "N"),
        (1, 0),
    ]

    def _match_bool_pattern(self, data: ndarray, pattern: Tuple) -> bool:
        match_positive: bool = False
        match_negative: bool = False

        for value in data:
            try:
                casted_value = self.cast_value(value, pattern)
            except ValueError:
                return False
            if casted_value is True:
                match_positive = True
            else:
                match_negative = True

        if match_positive and match_negative:
            return True
        return False

    def validate(self, data: ndarray) -> Optional[InferredField]:
        if data.dtype == np.dtype("bool"):
            return InferredField(inferred_type=self.name)
        for pattern in self.patterns:
            if self._match_bool_pattern(data, pattern):
                return InferredField(inferred_type=self.name, inferred_pattern=pattern)

    def cast_column(self, column: Series, inferred_field: InferredField) -> Series:
        raise NotImplementedError()

    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        if pattern is None:
            return value
        if isinstance(pattern[0], str) and isinstance(value, str):
            if self.is_empty_value(value):
                return None
            if value.lower() == pattern[0].lower():
                return True
            if value.lower() == pattern[1].lower():
                return False

        if isinstance(pattern[0], (bool, int)):
            if value == pattern[0]:
                return True
            if value == pattern[1]:
                return False

        raise ValueError(f"Could not cast value {value} using pattern {pattern}")

    def ignore_cast_for_types(self) -> Tuple[Any]:
        return (np.dtype("bool"),)
